--[[
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:
   
   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the following disclaimer
     in the documentation and/or other materials provided with the
     distribution.
   * Neither the name of the  nor the names of its
     contributors may be used to endorse or promote products derived from
     this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   
]]--

-- Same Game Remake
-- By Wesley "keyboard monkey" Werner
-- I ported some of this code from my Python samegame remake:
-- https://github.com/wesleywerner/mvc-game-design/tree/master/samegame

VERSION = "1"

-- Beautiful solarized colors
-- http://ethanschoonover.com/solarized
colors = {
    ["base02"] = {7, 54, 66},
    ["base00"] = {101, 123, 131},
    ["base01"] = {88, 110, 117},
    ["base1"] = {147, 161, 161},
    ["base2"] = {238, 232, 213},
    ["yellow"] = {181, 137, 0},
    ["orange"] = {203, 75, 22},
    ["red"] = {220, 50, 47},
    ["magenta"] = {211, 54, 130},
    ["violet"] = {108, 113, 196},
    ["blue"] = {38, 139, 210},
    ["cyan"] = {42, 161, 152},
    ["green"] = {133, 153, 0}
    }

colors.background = colors.base02
colors.selection = colors.base2

-- Defines the colors (and variation) of cells
cellColors = {colors.blue, colors.green, colors.magenta}

-- Initialize variables
defaultPlayTime = 90
grid = nil
blockSize = nil
testLevelCompleteCounter = 0
state = nil
playTime = nil
score = nil
targetScore = 0
scoreUpdateDelay = 0
soundOn = true
particleFlowTime = 0
highScores = { }
nameInput = ""

--                        _      _ 
--    _ __ ___   ___   __| | ___| |
--   | '_ ` _ \ / _ \ / _` |/ _ \ |
--   | | | | | | (_) | (_| |  __/ |
--   |_| |_| |_|\___/ \__,_|\___|_|

-- Returns a new 2-dimensional matrix. nil values assign random values.
function createMatrix (w, h, value)

    local g = { }
    for row = 1, h do
        g[row] = { }
        for col = 1, w do
            g[row][col] = value or math.random(1, #cellColors)
        end
    end
    return g

end

-- Returns a new game grid.
function createNewGrid (w, h)

    grid = { }
    grid.x = 10
    grid.y = 60
    grid.width = w
    grid.height = h
    grid.cells = createMatrix (w, h, nil)
    grid.selected = createMatrix (w, h, 0)
    grid.selectedCount = 0
    grid.blockCount = grid.width * grid.height

end

-- Select the block at xy and any surrounding blocks of the same value.
-- The grid.selected matrix gets a 1 for cells selected, and 0 for not selected.
function selectBlocks (row, col)

    -- Clear the selected matrix
    grid.selected = createMatrix (grid.width, grid.height, 0)
    grid.selectedCount = 0
    
    -- Abort invalid indices
    if (row == 0 or col == 0) then return end

    -- do not select zero values
    if grid.cells[row][col] == 0 then return end
    
    -- Keep a stack of cells to test
    local stack = { }
    
    -- Track cells already tested tested cells
    local tested = createMatrix (grid.width, grid.height, 0)
    
    grid.selectedCount = 1
    
    -- we use a local func to ease code reuse.
    local function mark_hotspots(row, col, n_row, n_col)
    
        -- test for a neighboring match
        if grid.cells[row][col] == grid.cells[n_row][n_col] then

            -- Does this cell need testing?
            if tested[n_row][n_col] == 0 then

                -- mark current cell as hot
                grid.selected[n_row][n_col] = 1
                grid.selectedCount = grid.selectedCount + 1

                -- add this cell to the stack so it's neighbors get tested later
                table.insert (stack, { ["row"] = n_row, ["col"] = n_col } )

                -- do not to check this cell in the future
                tested[n_row][n_col] = 1

            end
            
        end

    end

    -- starting cell is selected 
    grid.selected[row][col] = 1
    
    -- ... and tested
    tested[row][col] = 1
    
    -- ... and in the stack for testing (we need to start somewhere)
    table.insert (stack, { ["row"] = row, ["col"] = col } )
    
    -- while there are cells to test for friendly neighbors
    while #stack > 0 do
    
        -- test the four cardinal points around this cell.
        local p = table.remove (stack)
        local row, col = p.row, p.col
    
        -- west / left
        if (col > 1) then
            mark_hotspots(row, col, row, col - 1)
        end
    
        -- east / right
        if (col < grid.width) then
            mark_hotspots(row, col, row, col + 1)
        end
    
        -- north / above
        if (row > 1) then
            mark_hotspots(row, col, row - 1, col)
        end
    
        -- south / below
        if (row < grid.height) then
            mark_hotspots(row, col, row + 1, col)
        end
    
    end
    
end

-- Claim the selected cells by removing them from the grid
function claimSelected ( )

    local totalClaimed = 0

    -- Clear selected cells
    for row = 1, grid.height do
        for col = 1, grid.width do
            if grid.selected[row][col] == 1 then
                notify ('claimed', row, col, grid.cells[row][col])
                grid.cells[row][col] = 0
                totalClaimed = totalClaimed + 1
            end
        end
    end
    
    grid.selected = createMatrix (grid.width, grid.height, 0)
    grid.selectedCount = 0
    
    shiftDown ( )
    shiftLeft ( )
    
    -- Add score. More claimed yields higher rewards.
    targetScore = math.floor(math.log(totalClaimed+1)^6)
    
    -- Add to the time
    playTime = playTime + math.log(totalClaimed+1)^1
    
end

-- shift matrix rows down, from the bottom up, to fill empty spaces.
-- think of it as gravity making blocks fall into empty spaces below.
function shiftDown ( )

    for c = 1, grid.width do
    
        -- test from the bottom up
        r = grid.height - 1
        
        local lowestr = r
        
        while r > 0 do
        
            -- if this cell has a value, and the one below is empty
            if grid.cells[r][c] > 0 and grid.cells[r + 1][c] == 0 then
                -- notify of the drop
                --if (r < lowestr) then
                    --lowestr = r
                    --notify ('dropped', r, c, grid.cells[r][c], r + 1, c)
                --end
                -- switch their values
                grid.cells[r + 1][c] = grid.cells[r][c]
                grid.cells[r][c] = 0
                -- move the counter down, essentially testing the cell
                -- we just dropped, but in its new position.
                if r < grid.height - 1 then
                    r = r + 1
                end
            else
                --# nothing to move now. climb up the rows until we find
                --# a cell to drop, or we hit the top.
                r = r - 1
            end
        
        end
        
    end
    
end

-- shift matrix colums left to take up empty spaces.
function shiftLeft ( )

    local function isempty (col)
        -- test if a column has no values.
        for row = 1, grid.height do
            if grid.cells[row][col] > 0 then
                return false
            end
        end
        return true
    end
    
    local function copyleft (col)
        -- copy the column values across.
        for row = 1, grid.height do
            local value = grid.cells[row][col + 1]
            -- notify of the shift
            --if (value > 0) then
                --notify ('shifted', row, col + 1, value, row, col)
            --end
            grid.cells[row][col] = value
            grid.cells[row][col + 1] = 0
        end
    end
    
    -- test each column from the left
    local col = 1
    while col < grid.width do
        if (isempty(col) and not isempty(col + 1)) then
            copyleft(col)
            col = 1
        else
            col = col + 1
        end
    end
    
end

-- Up the game level and start a new round
function levelUp ()

    level = level + 1
    
    -- The difficulty increases logarithmically
    local upskill = math.log(level+1)^1.5
    local gridSize = math.floor(10 * upskill)
    
    createNewGrid (gridSize, gridSize)
    
    -- Constrain the block drawing size to fit the window
    blockSize = (love.window.getHeight() - (grid.y * 2)) / gridSize
    
    -- Reposition the grid to the centre of the screen
    grid.x = (love.window.getWidth() - (gridSize * blockSize)) / 2
    
    playTime = defaultPlayTime
    playTimeN = playTime
    
    notify ('leveled')

end
                
function testLevelComplete ( )

    -- The player has one or none selected already
    if (grid.selectedCount < 2) then
    
        for row = 1, grid.height do
            for col = 1, grid.width do
                selectBlocks (row, col)
                if (grid.selectedCount > 1) then
                    selectBlocks (0, 0)
                    return
                end
            end
        end
        
        -- Test for a perfect score
        if (cellsRemaining ( ) == 0) then
            love.audio.play (perfectSound)
            targetScore = targetScore + 500
        end
        
        levelUp ()

    end 
    
end

-- Cound how many cells are in the grid
function cellsRemaining ( )

    local count = 0
    for row = 1, grid.height do
        for col = 1, grid.width do
            if grid.cells[row][col] > 0 then
                count = count + 1
            end
        end
    end
    return count

end

function clickAt (x, y)

    -- Compensate for the grid position, and 1-based indices of the grid
    local row = math.floor((y - grid.y) / blockSize) + 1
    local col = math.floor((x - grid.x) / blockSize) + 1
    
    -- Sanity test
    if (row < 1 or col < 1 or row > grid.height or col > grid.width) then return end
    
    -- Clicking an existing selected cell will claim those cells
    local selected = grid.selected[row][col]
    if (selected == 1 and grid.selectedCount > 1) then
        -- claim
        claimSelected ()
        -- particles
        if (useParticles) then
            local mousex, mousey = love.mouse.getPosition ( )
            psystem:setPosition (mousex, mousey)
            psystem:start ( )
            particleFlowTime = 0.5
        end
        -- sound
        if soundOn then love.audio.play (claimSound) end
    else
        selectBlocks (row, col)
        if (grid.selectedCount > 0) then
            if soundOn then love.audio.play (selectSound) end
        end
    end

end

-- Receive notifications on game events
function notify (event, row, col, value, nrow, ncol)

    -- events:  dropped (row, col, value, nrow, ncol)
    --          claimed (row, col, value)
    --          shifted (row, col, value, nrow, ncol)
    
    -- A cell event has a row and col value
    if (row and col) then
        addAnimation (event, row, col, value, nrow, ncol)
    end
    
end

function newGame ( )

    nameInput = ""
    level = 0
    score = 0
    targetScore = 0
    state = "play"
    levelUp ()

end

function saveGame ( )

    if (state == "paused") then

        local binser = require "binser"

        local contents = binser.serialize(VERSION, level, score + targetScore, playTime, grid, blockSize)

        success = love.filesystem.write ("savegame.dat", contents, nil)
        
    elseif (state == "game over" or state == "main") then
    
        love.filesystem.remove ("savegame.dat")
        
    end

end

function loadGame ( )

    local is_file = love.filesystem.isFile ("savegame.dat")
    
    if (is_file) then
        
        contents, size = love.filesystem.read ("savegame.dat", nil)
        
        local fileVersion = nil
        local binser = require "binser"

        fileVersion, level, score, playTime, grid, blockSize = binser.deserialize(contents)
        
        if (fileVersion ~= VERSION) then
            print ("The saved game if from version " .. fileVersion .. ", and you are running version " .. VERSION)
            return
        end
        
        state = "paused"

    end
    
end

function saveScores ( )

    local binser = require "binser"

    local contents = binser.serialize(VERSION, highScores)

    success = love.filesystem.write ("scores.dat", contents, nil)

end

function loadScores ( )

    local is_file = love.filesystem.isFile ("scores.dat")
    
    if (is_file) then
        
        contents, size = love.filesystem.read ("scores.dat", nil)
        
        local fileVersion = nil
        local binser = require "binser"

        fileVersion, highScores = binser.deserialize(contents)
        
        if (fileVersion ~= VERSION) then
            print ("The scores file version " .. fileVersion .. ", and you are running version " .. VERSION)
        end
        
    else
    
        table.insert (highScores, { ["name"] = "Skeletor", ["score"] = 10000 })
        table.insert (highScores, { ["name"] = "Evil-Lyn", ["score"] = 9000 })
        table.insert (highScores, { ["name"] = "Panthor", ["score"] = 8000 })
        table.insert (highScores, { ["name"] = "Beast Man", ["score"] = 5000 })
        table.insert (highScores, { ["name"] = "Trap Jaw", ["score"] = 2500 })

    end

end

function addScore ( )

    table.insert (highScores, { ["name"] = nameInput, ["score"] = score })
    
    table.sort (highScores, function (a, b)
        return b.score < a.score
    end)
    
    while (#highScores > 10) do
        table.remove (highScores, 10)
    end
    
    saveScores ( )

end

--        _                    _             
--     __| |_ __ __ ___      _(_)_ __   __ _ 
--    / _` | '__/ _` \ \ /\ / / | '_ \ / _` |
--   | (_| | | | (_| |\ V  V /| | | | | (_| |
--    \__,_|_|  \__,_| \_/\_/ |_|_| |_|\__, |
--                                     |___/ 

function drawCell (row, col)

    local x = grid.x + (col * blockSize) - blockSize
    local y = grid.y + (row * blockSize) - blockSize
    local value = grid.cells[row][col]
    local selected = grid.selected[row][col]
    
    if (value > 0) then

        if selected == 1 then
            love.graphics.setColor (unpack (colors.selection))
        else
            love.graphics.setColor (unpack (cellColors[value]))
        end

        local scale = blockSize / 16
        love.graphics.draw (cellOverlay, x, y, 0, scale, scale)
        --love.graphics.rectangle ("fill", x, y, blockSize, blockSize)
        love.graphics.setLineWidth (2)
        love.graphics.setLineStyle ("rough")
        love.graphics.setColor (unpack (colors.base02))
        love.graphics.rectangle ("line", x, y, blockSize, blockSize)

    end
    
end

--               _                 _   _                 
--    __ _ _ __ (_)_ __ ___   __ _| |_(_) ___  _ __  ___ 
--   / _` | '_ \| | '_ ` _ \ / _` | __| |/ _ \| '_ \/ __|
--    (_| | | | | | | | | | | (_| | |_| | (_) | | | \__ \
--   \__,_|_| |_|_|_| |_| |_|\__,_|\__|_|\___/|_| |_|___/

animations = { }

function animationExists (row, col)

    return animations[row .. col] ~= nil

end

function addAnimation (effect, row, col, value, nrow, ncol)

    d = {
        ["effect"] = effect,
        ["row"] = row,
        ["col"] = col,
        ["value"] = value,
        ["nrow"] = nrow,
        ["ncol"] = ncol,
        ["ticks"] = 0,
        ["x"] = grid.x + (col * blockSize) - blockSize,
        ["y"] = grid.y + (row * blockSize) - blockSize,
        ["w"] = blockSize,
        ["h"] = blockSize
        }

    if (nrow and ncol) then
        d["dx"] = grid.x + (ncol * blockSize) - blockSize
        d["dy"] = grid.y + (nrow * blockSize) - blockSize
    end

    animations[row .. col] = d
    
end

function updateAnimations (dt)

    for k, a in pairs(animations) do
        if (a.effect == "claimed") then
            a.x = a.x + (blockSize / 2 * dt)
            a.w = a.w - (blockSize * dt)
            a.y = a.y + (blockSize / 2 * dt)
            a.h = a.h - (blockSize * dt)
            if (a.w <= 0) then
                animations[k] = nil
            end
        end
        if (a.effect == "dropped") then
            a.y = a.y + (100 * dt)
            if (a.y >= a.dy) then
                animations[k] = nil
            end
        end
        if (a.effect == "shifted") then
            a.x = a.x - (100 * dt)
            if (a.x <= a.dx) then
                animations[k] = nil
            end
        end
    end
    
end

function drawAnimations ( )

    for k, a in pairs(animations) do
        if (a.value > 0) then
        
            love.graphics.setColor (unpack (cellColors[a.value]))
            --love.graphics.rectangle ("fill", a.x, a.y, a.w, a.h)
            
            -- Draw the cell image scaled
            local scale = a.w / 16
            love.graphics.draw (cellOverlay, a.x, a.y, 0, scale, scale)

        end
    end
    
end

--    _                
--   | | _____   _____ 
--   | |/ _ \ \ / / _ \
--   | | (_) \ V /  __/
--   |_|\___/ \_/ \___|

function love.load ( )

    love.graphics.setDefaultFilter("nearest", "nearest", 1)
    smallFont = love.graphics.newFont ("images/Slackey.ttf", 16)
    largeFont = love.graphics.newFont ("images/Slackey.ttf", 60)
    math.randomseed (os.time ())
    love.graphics.setBackgroundColor ( unpack (colors["background"]) )
    love.graphics.setFont (smallFont)
    state = "main"
    cellOverlay = love.graphics.newImage ("images/glass.png")
    selectSound = love.audio.newSource ("sounds/select.wav", "static")
    claimSound = love.audio.newSource ("sounds/claim.wav", "static")
    scoreSound = love.audio.newSource ("sounds/score.wav", "static")
    timeSound = love.audio.newSource ("sounds/time.wav", "static")
    perfectSound = love.audio.newSource ("sounds/perfect.wav", "static")
    
    if (useParticles) then
        psystem = love.graphics.newParticleSystem (love.graphics.newImage ("images/star.png"), 10)
        psystem:setParticleLifetime(0.5, 4)
        psystem:setEmissionRate(5)
        psystem:setSizes(1, 10)
        psystem:setColors(255, 255, 255, 255, 255, 255, 255, 0) -- Fade to transparency.
        psystem:setSpeed(150, 300)
        psystem:setDirection(-4.71)
        psystem:stop ( )
    end
    
    loadGame ( )
    loadScores ( )
    
end

function love.mousepressed (x, y, button)

    if button == "l" then
    
        if (state == "play") then

            clickAt (x, y)
            
            -- Test if the level is complete in n seconds
            testLevelCompleteCounter = 1
    
        elseif (state == "main") then
        
            newGame ( )
            
        elseif (state == "paused") then
        
            state = "play"
            
        end
    
    end
    
    if (button == "r") then
        levelUp ()
    end

end

function love.keypressed (key, isrepeat)

    if (key == "f9") then
        state = "game over"
    end

    if (state == "game over") then
    
        if (key == "return") then
            addScore (nameInput)
            state = "main"
        else
            if (string.find("0123456789abcdefghijklmnopqrstuvwxyz ", key)) then
                nameInput = nameInput .. key
            elseif (key == "backspace") then
                nameInput = nameInput:sub(1, nameInput:len() - 1)
            end
        end
    
    elseif (key == "escape") then
    
        if (state == "main" or state == "game over") then
        
            love.event.quit ()
            
        elseif (state == "play") then
        
            state = "paused"
            
        elseif (state == "paused") then
        
            state = "play"
            
        end
    
    elseif (key == "q") then
    
        if (state == "paused") then
        
            love.event.quit ()
            
        end
        
    elseif (key == " ") then
    
        if (state == "main") then
            newGame ( )
        elseif (state == "paused") then
            state = "play"
        elseif (state == "game over") then
            state = "main"
        end
        
    elseif (key == "s") then
    
        if (state == "paused") then
            soundOn = (not soundOn and true or false)
            if soundOn then love.audio.play (selectSound) end
        end
        
    elseif (key == "f11") then
    
        local width, height, flags = love.window.getMode ( )
        if (flags.fullscreen) then
            flags.fullscreen = false
        else
            flags.fullscreen = true
        end
        love.window.setMode (defaultWindowWidth, defaultWindowHeight, flags)
        if (grid) then
            grid.x = (love.window.getWidth() - (grid.width * blockSize)) / 2
        end
        
    end

end
 
function love.update (dt)

    if (state == "play") then
    
        if (useParticles) then
            particleFlowTime = particleFlowTime - dt
            if (particleFlowTime < 0) then
                psystem:stop ( )
            end
            psystem:update(dt)
        end

        updateAnimations (dt)
        
        -- Test for level complete condition periodically
        testLevelCompleteCounter = testLevelCompleteCounter - dt
        if (testLevelCompleteCounter < 0) then
            testLevelCompleteCounter = 1000     -- a ridiculous wait
            testLevelComplete ()
        end
        
        -- Count down play time
        playTime = playTime - dt
        
        -- Bell on low time
        if (playTime < 5) then
            if (playTimeN ~= math.ceil(playTime)) then
                playTimeN = math.ceil(playTime)
                if soundOn then love.audio.play (timeSound) end
            end
        end
        
        -- Time has run out
        if (playTime < 0) then
            state = "game over"
        end
        
        if (targetScore > 0) then
            scoreUpdateDelay = scoreUpdateDelay - dt
            if (scoreUpdateDelay < 0) then
                scoreUpdateDelay = 0.02
                score = score + 1
                targetScore = targetScore - 1
                if soundOn then love.audio.play (scoreSound) end
            end
        end

    end
    
end

function drawCells ( )

    for row = 1, grid.height do
        for col = 1, grid.width do
            drawCell (row, col)
        end
    end

end

function drawHighScores (x, y)

    drawCenterText ("Winners", largeFont, 200, colors.base01)

    love.graphics.setFont (smallFont)

    for k, v in ipairs(highScores) do
        love.graphics.setColor(colors.base2)
        love.graphics.print(tostring(v.score), x, y)
        love.graphics.setColor(colors.green)
        love.graphics.print(v.name, x + 100, y)
        y = y + 25
    end

end

function drawCenterText (words, font, y, color)

    love.graphics.setFont (font)
    love.graphics.setColor (0, 0, 0, 192)
    love.graphics.printf (words, 3, y + 3, love.window.getWidth(), "center")
    love.graphics.setColor (color)
    love.graphics.printf (words, 0, y, love.window.getWidth(), "center")

end
 
function love.draw ( )

    if (state == "play") then

        drawCells ( )
        drawAnimations ( )
        drawGameStats ( )

        if (useParticles) then
            love.graphics.draw(psystem, 0, 0)
        end

    elseif (state == "main") then
    
        drawCenterText ("Same Stones", largeFont, 10, colors.blue)
        
        -- Instructions
        drawCenterText ("Click to remove groups of same-colored stones. More stones in a group give more points, and watch your time!", smallFont, 100, colors.base2)
        
        -- Created by
        drawCenterText ("Created in LÖVE", smallFont, 550, colors.blue)
        drawCenterText ("by Wesley \"keyboard monkey\" Werner", smallFont, 570, colors.green)
        
        -- Draw cell images
        local scale = 96 / 16
        love.graphics.setColor (unpack (colors.red))
        love.graphics.draw (cellOverlay, 100, 300, 0, scale, scale)
        love.graphics.setColor (unpack (colors.green))
        love.graphics.draw (cellOverlay, 140, 340, 0, scale, scale)
        
        -- Sccores
        drawHighScores (160, 280)

    elseif (state == "game over") then
    
        drawCells ( )
        drawGameStats ( )
        drawCenterText ("Game Over", largeFont, 100, colors.base2)
        drawCenterText ("Score: " .. tostring(score), largeFont, 220, colors.green)
        drawCenterText ("Enter your name:", smallFont, 360, colors.base2)
        drawCenterText (nameInput, largeFont, 400, colors.magenta)

    elseif (state == "paused") then
    
        drawCenterText ("Paused", largeFont, 100, colors.green)
        drawCenterText ("[Q]uit", smallFont, 240, colors.base2)
        drawCenterText ("[S]ound is " .. (soundOn and "on" or "off"), smallFont, 280, colors.base2)
        drawCenterText ("F11 Fullscreen", smallFont, 320, colors.base2)
        drawGameStats ( )

    end
    
end

function love.quit ( )

    saveGame ( )

end

function drawGameStats ( )

    love.graphics.setFont (smallFont)
    
    local x, y = grid.x + 10, grid.y + (grid.height * blockSize)

    love.graphics.setColor (unpack (colors.blue))
    love.graphics.print ("Level", x, y)
    love.graphics.setColor (unpack (colors.base1))
    love.graphics.print (tostring(level), x + 20, y + 20)
    
    x = x + 100
    love.graphics.setColor (unpack (colors.blue))
    love.graphics.print ("Blocks", x, y)
    love.graphics.setColor (unpack (colors.base1))
    love.graphics.print (tostring(grid.blockCount), x + 20, y + 20)

    x = x + 100
    love.graphics.setColor (unpack (colors.blue))
    love.graphics.print ("Score ", x, y)
    love.graphics.setColor (unpack (colors.base1))
    love.graphics.print (tostring(score), x + 20, y + 20)

    x = x + 100
    love.graphics.setColor (unpack (colors.blue))
    love.graphics.print ("Selected ", x, y)
    love.graphics.setColor (unpack (colors.base1))
    love.graphics.print (tostring(grid.selectedCount), x + 20, y + 20)

    if (state == "play" or state == "paused") then
        x = x + 100
        if (playTime < 15) then
            love.graphics.setColor (unpack (colors.red))
        else
            love.graphics.setColor (unpack (colors.blue))
        end
        love.graphics.print ("Time ", x, y)
        love.graphics.setColor (unpack (colors.yellow))
        love.graphics.print (tostring(math.ceil(playTime)), x + 20, y + 20)
    end
    
end

--        _      _                 
--     __| | ___| |__  _   _  __ _ 
--    / _` |/ _ \ '_ \| | | |/ _` |
--   | (_| |  __/ |_) | |_| | (_| |
--    \__,_|\___|_.__/ \__,_|\__, |
--                           |___/ 

function printGrid ( )

    print ('blocks grid')
    for i = 1, grid.height do
        local s = ""
        for j = 1, grid.width do
            s = s .. grid.cells[i][j]
        end
        print(s)
    end

end

function printSelected ( )

    print ('selected blocks')
    for i = 1, grid.height do
        local s = ""
        for j = 1, grid.width do
            s = s .. grid.selected[i][j]
        end
        print(s)
    end

end
